# Ensure Python 2/3 compatibility: http://python-future.org/compatible_idioms.html
from __future__ import print_function
from future.utils import iteritems

"""
.. module:: configurator
    :platform: Win/Linux/Mac
    :synopsis: This module parses config files and retrieves data
    :plans: Support beyond 1 level nesting and customization apart from my current setup
    :changelog:
        2.0 - Added YAML support
"""

import yaml
import os
from collections import OrderedDict
from six import add_metaclass
from pprint import pformat
import nomenclate.core.exceptions as exceptions
from nomenclate.core.tools import (
    gen_dict_key_matches,
    get_keys_containing
)
from nomenclate.core.nlog import (
    getLogger,
    DEBUG,
    INFO,
    CRITICAL
)


class ConfigParse(object):
    LOG = getLogger(__name__, level=CRITICAL)

    def __init__(self, config_filepath='env.yml'):
        """
        Args:
            section (str): section to query
            subsection (str):subsection to query
        """
        self.config_filepath = self.get_default_config_file(config_filepath)
        self.config_file_contents = None
        self.config_entry_handler = ConfigEntryFormatter()
        self.rebuild_config_cache(self.config_filepath)

    def get_default_config_file(self, config_filepath):
        try:
            self.validate_config_file(config_filepath)
            return config_filepath
        except (IOError, OSError):
            pass

        try:
            cwd_filepath = os.path.normpath(os.path.join(os.getcwd(), config_filepath))
            cur_file_filepath = os.path.normpath(os.path.join(os.path.dirname(__file__), config_filepath))
            return cwd_filepath if os.path.isfile(cwd_filepath) else cur_file_filepath
        except IOError:
            raise exceptions.SourceError('No config file found in current working directory or nomenclate/core')

    def rebuild_config_cache(self, config_filepath):
        """ Loads from file and caches all data from the config file in the form of an OrderedDict to self.data
        Args:
            config_filepath (str): the full filepath to the config file
        Returns (bool): success status
        """
        self.validate_config_file(config_filepath)
        config_data = None
        try:
            with open(config_filepath, 'r') as f:
                config_data = yaml.load(f)
            items = list(iteritems(config_data))

        except AttributeError:
            items = list(config_data)

        self.config_file_contents = OrderedDict(sorted(items, key=lambda x: x[0], reverse=True))
        self.config_filepath = config_filepath

    def get(self, query_path, return_type=list, preceding_depth=None):
        """ Traverses the list of query paths to find the data requested
        Args:
            query_path list(str)]: list of query path branches
            return_type (Union[list, str, dict, OrderedDict]): desired return type for the data
            preceding_depth (int): always returns a dictionary encapsulating the data
                                    that traces back up the path for x depth
                                    -1 for the full traversal back up the path
                                    None is default for no traversal
        Raises:
            exceptions.ResourceNotFoundError: if the query path is invalid
        """
        function_type_lookup = {str: self._get_path_entry_from_string,
                                list: self._get_path_entry_from_list}

        self.LOG.debug('config.get() - Trying to find %s in config and return_type %s' % (repr(query_path), return_type))

        if not query_path:
            return self._default_config(return_type)

        try:
            config_entry = function_type_lookup[type(query_path)](query_path)
            self.LOG.debug('Retrieved config entry:\n%s' % pformat(config_entry, depth=1))
            query_result = self.config_entry_handler.format_query_result(config_entry,
                                                                         query_path,
                                                                         return_type=return_type,
                                                                         preceding_depth=preceding_depth)
            self.LOG.debug('Converted config entry:\n%s' % pformat(query_result, depth=1))
            return query_result
        except IndexError:
            self.LOG.debug('Not sure why but IndexError was found...defaulting to return: %s = %r' %
                           (return_type, return_type()))
            return return_type()

    def _get_path_entry_from_string(self, query_string, first_found=True, full_path=False):
        iter_matches = gen_dict_key_matches(query_string, self.config_file_contents, full_path=full_path)
        try:
            return next(iter_matches) if first_found else iter_matches
        except (StopIteration, TypeError):
            raise exceptions.ResourceNotFoundError('Could not find search string %s in the config file contents %s' %
                                                   (query_string, self.config_file_contents))

    def _get_path_entry_from_list(self, query_path):
        cur_data = self.config_file_contents
        try:
            self.LOG.debug('starting path search from list...' % query_path)
            for child in query_path:
                self.LOG.debug(' -> %s' % child)
                cur_data = cur_data[child]
            self.LOG.debug('Found data %s' % cur_data)
            return cur_data
        except (AttributeError, KeyError):
            raise exceptions.ResourceNotFoundError('Could not find query path %s in the config file contents' %
                                                   query_path)

    def _default_config(self, return_type):
        self.LOG.debug('Returning default for type %s -> %s' % (return_type, repr(return_type())))
        if return_type == list:
            return [k for k in self.config_file_contents]
        return return_type()

    @classmethod
    def validate_config_file(cls, config_filepath):
        if not os.path.isfile(config_filepath) and os.path.isabs(config_filepath):
            raise IOError('File path %s is not a valid yml, ini or cfg file or does not exist' % config_filepath)

        elif os.path.getsize(config_filepath) == 0:
            raise IOError('File %s is empty' % config_filepath)

        with open(config_filepath, 'r') as f:
            if yaml.load(f) is None:
                raise IOError('No YAML config was found in file %s' % config_filepath)


class FormatterRegistry(type):
    CONVERSION_TABLE = {}

    def __new__(mcs, name, bases, dct):
        cls = type.__new__(mcs, name, bases, dct)

        extensions = dct.get('converts')
        accepted_input_type = extensions.get('accepted_input_type', None)
        accepted_return_type = extensions.get('accepted_return_type', None)

        if accepted_input_type and accepted_return_type:
            take_exists = mcs.CONVERSION_TABLE.get(accepted_input_type)
            if not take_exists:
                mcs.CONVERSION_TABLE[accepted_input_type] = {}

            mcs.CONVERSION_TABLE[accepted_input_type][accepted_return_type] = cls

        return cls

    @classmethod
    def get_by_take_and_return_type(mcs, input_type, return_type):
        return mcs.CONVERSION_TABLE[input_type][return_type]


@add_metaclass(FormatterRegistry)
class BaseFormatter(object):
    converts = {'accepted_input_type': None,
                'accepted_return_type': None}

    def __init__(self, parent):
        self.parent = parent

    @staticmethod
    def format_result(input):
        raise NotImplementedError


class StringToListEntryFormatter(BaseFormatter):
    converts = {'accepted_input_type': str,
                'accepted_return_type': list}

    @staticmethod
    def format_result(input):
        return input.split()


class DictToStringEntryFormatter(BaseFormatter):
    converts = {'accepted_input_type': dict,
                'accepted_return_type': str}

    @staticmethod
    def format_result(input):
        return ' '.join(list(input))


class DictToListEntryFormatter(BaseFormatter):
    converts = {'accepted_input_type': dict,
                'accepted_return_type': list}

    @staticmethod
    def format_result(input):
        """ Always sorted for order
        """
        keys = list(input)
        keys.sort()
        return keys


class DictToOrderedDictEntryFormatter(BaseFormatter):
    converts = {'accepted_input_type': dict,
                'accepted_return_type': OrderedDict}

    @staticmethod
    def format_result(input):
        """From: http://stackoverflow.com/questions/13062300/convert-a-dict-to-sorted-dict-in-python
        """
        try:
            items = list(iteritems(input))
        except AttributeError:
            items = list(input.items())
        return OrderedDict(sorted(items, key=lambda x: x[0]))


class OrderedDictToListEntryFormatter(BaseFormatter):
    converts = {'accepted_input_type': OrderedDict,
                'accepted_return_type': list}

    @staticmethod
    def format_result(input):
        """From: http://stackoverflow.com/questions/13062300/convert-a-dict-to-sorted-dict-in-python
        """
        return list(input)


class NoneToDictEntryFormatter(BaseFormatter):
    converts = {'accepted_input_type': type(None),
                'accepted_return_type': dict}

    @staticmethod
    def format_result(input):
        """From: http://stackoverflow.com/questions/13062300/convert-a-dict-to-sorted-dict-in-python
        """
        return {}


class ListToStringEntryFormatter(BaseFormatter):
    converts = {'accepted_input_type': list,
                'accepted_return_type': str}

    @staticmethod
    def format_result(input):
        return ' '.join(input)


class IntToListEntryFormatter(BaseFormatter):
    converts = {'accepted_input_type': int,
                'accepted_return_type': list}

    @staticmethod
    def format_result(input):
        return [input]


class ConfigEntryFormatter(object):
    def format_query_result(self, query_result, query_path, return_type=list, preceding_depth=None):
        """
        Args:
            query_result (dict|str|list): yaml query result
            query_path [str]: list of strings representing query path
            return_type (rtype): return type of object user desires
            preceding_depth (int): the depth to which we want to encapsulate back up config tree
                                    -1 : defaults to entire tree
        Returns (dict|OrderedDict|str|list): specified return type
        """
        if type(query_result) != return_type:
            converted_result = self.format_with_handler(query_result, return_type)
        else:
            converted_result = query_result

        converted_result = self.add_preceding_dict(converted_result, query_path, preceding_depth)
        return converted_result

    def get_handler(self, query_result_type, return_type):
        try:
            return FormatterRegistry.get_by_take_and_return_type(query_result_type, return_type)
        except AttributeError:
            raise AttributeError(
                "Handler not found for return type %s and input type %s" % (return_type, type(query_result_type)))
        except:
            msg = 'Could not find function in conversion list for input type %s and return type %s' % \
                  (query_result_type, return_type)
            raise IndexError(msg)

    def format_with_handler(self, query_result, return_type):
        handler = self.get_handler(type(query_result), return_type)
        return handler.format_result(query_result)

    def add_preceding_dict(self, config_entry, query_path, preceding_depth):
        if preceding_depth is None:
            return config_entry

        preceding_dict = {query_path[-1]: config_entry}
        path_length_minus_query_pos = len(query_path) - 1
        preceding_depth = path_length_minus_query_pos - preceding_depth if preceding_depth != -1 else 0

        for index in reversed(range(preceding_depth, path_length_minus_query_pos)):
            preceding_dict = {query_path[index]: preceding_dict}

        return preceding_dict



#!/usr/bin/env python
# Ensure Python 2/3 compatibility: http://python-future.org/compatible_idioms.html
from __future__ import print_function

from imp import reload

"""
.. module:: nameparser
    :platform: Win/Linux/Mac
    :synopsis: This module does name parsing to recover information about the name based on common conventions.
                it will be very prone to missing information since names will be very variable, but hopefully
                it will be more intelligent as time goes on.
    :plans:
        0.1.0: Added base template functionality
        0.2.0: Added abbreviation generation technology
        0.2.1: Removed nomenclate.core.toolbox opting to auto-configure config path
               Optimized get_side to use the env.ini settings
        0.3.0: Normalized the returns of each of the get functions and added an extra function to parse the entire name
               Fixed testing for everything as well.
               All static data stored as class variables now
               Need to implement both get_discipline and get_initials still
"""
from pprint import pprint
import re
import datetime
import itertools
from future.utils import iteritems
import nomenclate.core.configurator as config

__author__ = "Andres Weber"
__email__ = "andresmweber@gmail.com"
__version__ = '0.3.0'

reload(config)


class NameParser(object):
    """ This parses names of assets.  It assumes the usual convention of strings separated by underscores.
    """
    CONFIG_SIDES = config.ConfigParse().get(['options', 'side'])
    CONFIG_DISCIPLINES = config.ConfigParse().get(['options', 'discipline'])
    PARSABLE = ['basename', 'version', 'date', 'side', 'udim']

    REGEX_BASENAME = r'(?:^[-._]+)?([a-zA-Z0-9_\-|]+?)(?=[-._]{2,}|\.)'
    REGEX_SEPARATORS = r'[ ,_!?:\.\-]'

    REGEX_ABBR_ISLAND = r'(?:[{SEP}]+)({ABBR})(?:[{SEP}]+)'
    REGEX_ABBR_SEOS = r'(?:^|[a-z]|{SEP})({ABBR})(?:$|[A-Z][a-z]+|{SEP})'
    REGEX_ABBR_CAMEL = r'(?:[a-z])({ABBR})(?:$|[A-Z]|{SEP})'

    REGEX_VERSION = r'(?:^|[a-z]{{2,}}[a-uw-z]+|{SEP}[a-z]?[a-uw-z]+|{SEP})([vV]?[0-9]{{1,4}})(?={SEP}|$)'
    REGEX_UDIM = r'(?:[a-zA-Z]|[._-])(1[0-9]{3})(?:[._-])'
    REGEX_DATE = r'(?<!\d)(%s)(?!\d)'
    REGEX_CAMEL = r'(?:{SEP}?)((?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z]))(?:{SEP}?)'

    @classmethod
    def parse_name(cls, name):
        """ Parses a name into a dictionary of identified subsections with accompanying information to
            correctly identify and replace if necessary
        Args:
            name (str): string to be parsed
        Returns (dict): dictionary with relevant parsed information
        """
        parse_dict = dict.fromkeys(cls.PARSABLE, None)
        parse_dict['date'] = cls.get_date(name)
        parse_dict['version'] = cls.get_version(name)
        parse_dict['udim'] = cls.get_udim(name)
        parse_dict['side'] = cls.get_side(name)
        parse_dict['basename'] = cls.get_base_naive(cls._reduce_name(name, parse_dict))
        return parse_dict

    @classmethod
    def get_side(cls, name, ignore=''):
        """ Checks a string for a possible side string token, this assumes its on its own
            and is not part of or camel cased and combined with a word.  Returns first found side to reduce duplicates.
            We can be safe to assume the abbreviation for the side does not have camel casing within its own word.
        Args:
            name (str): string that represents a possible name of an object
        """
        for side in cls.CONFIG_SIDES:
            """ Tried using a regex, however it would've taken too long to debug
            side_regex = cls._build_abbreviation_regex(side)
            print ('side regex for side %s is %s'%(side, side_regex))
            result = cls._generic_search(name, side_regex, metadata={'side': side}, ignore=ignore)
            if result:
                print ('found a result, coming back.')
                return result
            """
            for permutations in cls.get_string_camel_patterns(side):
                for permutation in permutations:
                    result = cls._generic_search(name, permutation, metadata={'side': side}, ignore=ignore)
                    if result:
                        return result
        return None

    @classmethod
    def get_discipline(cls, name, ignore='', min_length=3):
        """
        Args:
            name (str): the string based object name
            ignore (str): specific ignore string for the search to avoid
            min_length (int): minimum length for possible abbrevations of disciplines.  Lower = more wrong guesses.
        Returns (dict): match dictionary
        """
        for discipline in cls.CONFIG_DISCIPLINES:
            re_abbr = '({RECURSE}(?=[0-9]|[A-Z]|{SEPARATORS}))'.format(
                RECURSE=cls._build_abbreviation_regex(discipline),
                SEPARATORS=cls.REGEX_SEPARATORS)
            matches = cls._get_regex_search(name, re_abbr, ignore=ignore)
            if matches:
                matches = [m for m in matches if
                           re.findall('([a-z]{%d,})' % min_length, m['match'], flags=re.IGNORECASE)]
                if matches:
                    return matches[-1]
        return None

    @classmethod
    def get_base(cls, name):
        """ Checks a string for a possible base name of an object (no prefix, no suffix).
            We need to do a full parse to make sure we've ruled out all the possible other parse queries
        Args:
            name (str): string that represents a possible name of an object
        Returns (str): the detected basename
        """
        return cls.parse_name(name).get('basename', None)

    @classmethod
    def get_base_naive(cls, name, ignore=''):
        """ Checks a string for a possible base name of an object (no prefix, no suffix).
            We need to do a full parse to make sure we've ruled out all the possible other parse queries
        Args:
            name (str): string that represents a possible name of an object
        Returns (str): the detected basename
        """
        return cls._get_regex_search(name, cls.REGEX_BASENAME, match_index=0, ignore=ignore)

    @classmethod
    def get_version(cls, name):
        """ Checks a string for a possible version of an object (no prefix, no suffix).
            Assumes only up to 4 digit padding
        Args:
            name (str): string that represents a possible name of an object
        Returns (float|int, [str]): gets the version number as a float or int if whole then the string matches or None
        """
        # Dates can confuse th
        # is stuff, so we'll check for that first and remove it from the string if found
        try:
            date = cls.get_date(name)
            date = date['datetime'].strftime(date['format'])
        except TypeError:
            pass
        return cls.get_version_naive(name, ignore=date or '')

    @classmethod
    def get_version_naive(cls, name, ignore=''):
        """ Checks a string for a possible version of an object (no prefix, no suffix) without filtering date out
            Assumes only up to 4 digit padding
        Args:
            name (str): string that represents a possible name of an object
        Returns (float|int, [str]): gets the version number as a float or int if whole then the string matches or None
        """
        match = cls._get_regex_search(name, cls.REGEX_VERSION.format(SEP=cls.REGEX_SEPARATORS), ignore=ignore)

        if match is not None:
            if len(match) > 1:
                for m in match:
                    m.update({'version': int(m['match'].upper().replace('V', ''))})
                compound_version = '.'.join([str(m['version']) for m in match])
                compound_version = float(compound_version) if compound_version.count('.') == 1 else compound_version
                return {'compound_matches': match,
                        'compound_version': compound_version,
                        'pattern': match[0]['pattern'],
                        'input': match[0]['input']}

            elif len(match) == 1:
                match = match[0]
                match.update({'version': int(match['match'].upper().replace('V', ''))})
                return match

        return None

    @classmethod
    def get_udim(cls, name):
        """ Checks a string for a possible base name of an object (no prefix, no suffix)
        Args:
            name (str): string that represents a possible name of an object
        Returns (int): the last found match because convention keeps UDIM markers at the end.
        """
        match = cls._get_regex_search(name, cls.REGEX_UDIM, match_index=-1)
        if match:
            match.update({'match_int': int(match['match'])})
            return match
        return None

    @classmethod
    def get_short(cls, name):
        """ Returns the short name of a Maya asset name or path
        Args:
            name (str): string that represents a possible name of an object
        """
        return name.split('|')[-1].split('//')[-1]

    @classmethod
    def get_date(cls, name):
        """ Checks a string for a possible date formatted into the name.  It assumes dates do not have other
            numbers at the front or head of the date.  Currently only supports dates in 1900 and 2000.
            Heavily relies on datetime for error checking to see
            if date is actually viable. It follows similar ideas to this post:
            http://stackoverflow.com/questions/9978534/match-dates-using-python-regular-expressions
            Args:
                name (str): string that represents a possible name of an object
            Returns (datetime.datetime): datetime object with current time or None if not found
        """
        time_formats = ['%Y-%m-%d_%H-%M-%S',
                        '%Y-%m-%d-%H-%M-%S',
                        '%Y-%m-%d--%H-%M-%S',
                        '%y_%m_%dT%H_%M_%S',
                        '%Y-%m-%d%H-%M-%S',
                        '%Y%m%d-%H%M%S',
                        '%Y%m%d-%H%M',
                        '%Y-%m-%d',
                        '%Y%m%d',
                        '%m_%d_%Y',
                        '%m_%d_%y',
                        '%m%d%y',
                        '%m%d%Y',
                        '%d_%m_%Y',
                        '%Y',
                        '%m-%d-%yy',
                        '%m%d%Y']

        mapping = [('%yy', '(([01]\d{1}))'), ('%Y', '((19|20)\d{2})'), ('%y', '(\d{2})'), ('%d', '(\d{2})'),
                   ('%m', '(\d{2})'),
                   ('%H', '(\d{2})'), ('%M', '(\d{2})'), ('%S', '(\d{2})')]
        time_regexes = []
        for time_format in time_formats:
            for k, v in mapping:
                time_format = time_format.replace(k, v)
            time_regexes.append(time_format)

        for time_regex, time_format in zip(time_regexes, time_formats):
            match = cls._get_regex_search(name,
                                          cls.REGEX_DATE % time_regex,
                                          metadata={'format': time_format},
                                          match_index=0)
            if match:
                try:
                    match.update({
                        'datetime': datetime.datetime.strptime(match['match'], time_format.replace('%yy', '%y'))
                        })
                    return match
                except ValueError:
                    pass
        return None

    @classmethod
    def get_string_camel_patterns(cls, name, min_length=0):
        """
        Args:
            name (str): the name we need to get all possible permutations and abbreviations for
            min_length (int): minimum length we want for abbreviations
        Return list(list(str)): list casing permutations of list of abbreviations
        """
        # Have to check for longest first and remove duplicates
        patterns = []
        abbreviations = list(set(cls._get_abbrs(name, output_length=min_length)))
        abbreviations.sort(key=len, reverse=True)

        for abbr in abbreviations:
            # We won't check for abbreviations that are stupid eg something with apparent camel casing within
            # the word itself like LeF, sorting from:
            # http://stackoverflow.com/questions/13954841/python-sort-upper-case-and-lower-case
            casing_permutations = list(set(cls._get_casing_permutations(abbr)))
            casing_permutations.sort(key=lambda v: (v.upper(), v[0].islower(), len(v)))
            permutations = [permutation for permutation in casing_permutations if
                            cls.is_valid_camel(permutation) or len(permutation) <= 2]
            if permutations:
                patterns.append(permutations)

        return patterns

    @classmethod
    def _reduce_name(cls, name, parse_dict):
        """ Reduces a name against matches found in a parse dictionary
        Args:
            name (str): name to be reduced
            parse_dict (dict): dictionary of matches to reduce against
        Returns (str): reduced string
        """
        # Now remove all found entries to make basename regex have an easier time
        removal_indices = []
        for section, match in iteritems(parse_dict):
            try:
                matches = []
                if isinstance(match, dict) and 'compound_matches' in match:
                    matches = match.get('compound_matches')
                elif not isinstance(match, list) and match is not None:
                    matches = [match]

                for m in matches:
                    valid_slice = True
                    slice_a, slice_b = m.get('position')
                    # Adjust slice positions from previous slices
                    if removal_indices is []:
                        removal_indices.append((slice_a, slice_b))

                    for r_slice_a, r_slice_b in removal_indices:
                        if slice_a == r_slice_a and slice_b == r_slice_b:
                            valid_slice = False
                        if slice_a > r_slice_a or slice_a > r_slice_b or slice_b > r_slice_b or slice_b > r_slice_a:
                            slice_delta = r_slice_b - r_slice_a
                            slice_a -= slice_delta
                            slice_b -= slice_delta

                    if valid_slice:
                        name = cls._string_remove_slice(name, slice_a, slice_b)
                        removal_indices.append((slice_a, slice_b))
            except (IndexError, TypeError):
                pass
        return name

    @staticmethod
    def _get_regex_search(input_string, regex, metadata={}, match_index=None, ignore='', flags=0):
        """ Using this so that all results from the functions return similar results
        Args:
            input_string (str): input string to be checked
            regex (str): input regex to be compiled and searched with
            match_index (int or None): whether to get a specific match, if None returns all matches as list
            metadata (dict): dictionary of extra metatags needed to identify information
        Returns list(dict): list of dictionaries if multiple hits or a specific entry or None
        """
        generator = re.compile(regex, flags=flags).finditer(input_string)
        matches = []
        for obj in generator:
            try:
                span_a = obj.span(1)
                group_a = obj.group(1)
            except IndexError:
                span_a = obj.span()
                group_a = obj.group()

            if obj.groups() == ('',):
                # Not sure how to account for this situation yet, weird regex.
                return True

            if group_a not in ignore:
                matches.append({'pattern': regex,
                                'input': input_string,
                                'position': span_a,
                                'position_full': obj.span(),
                                'match': group_a,
                                'match_full': obj.group()})

        if matches:
            for match in matches:
                match.update(metadata)
            if match_index is not None:
                return matches[match_index]
            return matches
        return None

    @classmethod
    def _generic_search(cls, name, search_string, metadata={}, ignore=''):
        """ Searches for a specific string given three types of regex search types.  Also auto-checks for camel casing.
        Args:
            name (str): name of object in question
            search_string (str): string to find and insert into the search regexes
            metadata (dict): metadata to add to the result if we find a match
            ignore (str): ignore specific string for the search
        Returns Optional(dict): dictionary of search results
        """
        patterns = [cls.REGEX_ABBR_SEOS,
                    cls.REGEX_ABBR_ISLAND,
                    cls.REGEX_ABBR_CAMEL]

        if not search_string[0].isupper():
            patterns.remove(cls.REGEX_ABBR_CAMEL)

        for pattern in patterns:
            search_result = cls._get_regex_search(name,
                                                  pattern.format(ABBR=search_string, SEP=cls.REGEX_SEPARATORS),
                                                  metadata=metadata,
                                                  match_index=0,
                                                  ignore=ignore)
            if search_result is not None:
                if cls.is_valid_camel(search_result.get('match_full'), strcmp=search_result.get('match')):
                    return search_result
        return None

    @staticmethod
    def _get_abbrs(input_string, output_length=0):
        """ Generates abbreviations for input_string
        Args:
            input_string (str): name of object
            output_length (num): optional specific length of abbreviations, default is off
        Returns: [str]: list of all combinations that include the first letter (possible abbreviations)
        """
        for i, j in itertools.combinations(range(len(input_string[1:]) + 1), 2):
            abbr = input_string[0] + input_string[1:][i:j]
            if len(abbr) >= output_length:
                yield abbr
            elif output_length == 0:
                yield abbr
        # Have to add the solitary letter as well
        if not output_length or output_length == 1:
            yield input_string[0]

    @classmethod
    def is_valid_camel(cls, input_string, strcmp=None, ignore=''):
        """ Checks to see if an input string is valid for use in camel casing
            This assumes that all lowercase strings are not valid camel case situations and no camel string
            can just be a capitalized word.  Took ideas from here:
                http://stackoverflow.com/questions/29916065/how-to-do-camelcase-split-in-python
        Args:
            input_string (str): input word
            strcmp (str): force detection on a substring just in case its undetectable
                          (e.g. part of a section of text that's all lowercase)
        Returns (bool): whether it is valid or not
        """
        # clear any non chars from the string
        if not input_string:
            return False

        input_string = ''.join([c for c in input_string if c.isalpha()])
        matches = cls._get_regex_search(input_string,
                                        cls.REGEX_CAMEL.format(SEP=cls.REGEX_SEPARATORS),
                                        match_index=0,
                                        ignore=ignore)
        if matches or input_string == strcmp:
            if strcmp:
                index = input_string.find(strcmp) - 1
                is_camel = strcmp[0].isupper() and input_string[index].islower()
                is_input = strcmp == input_string
                is_start = index + 1 == 0
                return is_camel or is_input or is_start
            return True
        elif len(input_string) == 1:
            return True
        return False

    @staticmethod
    def _split_camel(name):
        return re.sub('(?!^)([A-Z][a-z]+)', r' \1', name).split()

    @classmethod
    def _get_casing_permutations(cls, input_string):
        """ Takes a string and gives all possible permutations of casing for comparative purposes
        Args:
            input_string (str): name of object
        Yields: (str): iterator of all possible permutations of casing for the input_string
        """
        if not input_string:
            yield ""
        else:
            first = input_string[:1]
            for sub_casing in cls._get_casing_permutations(input_string[1:]):
                yield first.lower() + sub_casing
                yield first.upper() + sub_casing

    @staticmethod
    def _string_remove_slice(input_str, start, end, ref_string=''):
        """
        Args:
            input_str (str): input string
            start (int): end search index
            end (int): start search index
            ref_string (str): reference string for correct index references
        :return:
        """
        if 0 <= start < end <= len(input_str):
            return input_str[:start] + input_str[end:]
        return input_str

    @staticmethod
    def _build_abbreviation_regex(input_string):
        """ builds a recursive regex based on an input string to find possible abbreviations more simply.
            e.g. = punct(u(a(t(i(on?)?)?)?)?)?
        Args:
            input_string (str): input string
        Returns (str): output regex
        """
        result = '([%s%s]' % (input_string[0].upper(), input_string[0].lower())
        for char in input_string[1:]:
            result += '[%s%s]?' % (char.upper(), char.lower())
        return result + ')'



from future.utils import iteritems
import collections
from pprint import pformat
from nomenclate.core.nlog import (
    getLogger,
    DEBUG,
    INFO,
    CRITICAL
)

LOG = getLogger(__name__, level=CRITICAL)


def combine_dicts(*args, **kwargs):
    dicts = [arg for arg in args if isinstance(arg, dict)]
    LOG.info('dicts are %s' % pformat(dicts))
    dicts.append(kwargs)
    super_dict = collections.defaultdict(dict)

    for d in dicts:
        for k, v in iteritems(d):
            if k:
                super_dict[k] = v
    LOG.info('super dict is %s' % pformat(dict(super_dict)))
    return dict(super_dict)


def get_keys_containing(search_string, input_dict, default=None, first_found=True):
        output = {}
        for k, v in iteritems(input_dict):
            if search_string in k and not callable(k):
                output[k] = v

        if first_found:
            try:
                output = output[next(iter(output))]
            except StopIteration:
                pass

        output = output or default

        return output


def gen_dict_key_matches(key, dictionary, path=None, full_path=False):
    if path is None:
        path = []
    LOG.info('\nThe main input to the function is:\n %s\n' % pformat(dict(dictionary)))
    for k, v in iteritems(dictionary):
        path.append(k)
        if k == key:
            LOG.debug('\n\t\tkey %s matches query string %s, yielding!' % (k, key))
            yield (path, v) if full_path else v
        elif isinstance(v, dict):
            LOG.debug('\n\t\tvalue is a dictionary, iterating through %s!' % pformat(v))
            for result in gen_dict_key_matches(key, v, path):
                yield result


def flatten(it):
    """ From:
    http://stackoverflow.com/questions/11503065/python-function-to-flatten-generator-containing-another-generator
    """
    for x in it:
        if isinstance(x, collections.Iterable) and not isinstance(x, str):
            for y in flatten(x):
                yield y
        else:
            yield x



#!/usr/bin/env python
# Ensure Python 2/3 compatibility: http://python-future.org/compatible_idioms.html
from __future__ import print_function
from future.utils import iteritems
import re
import nomenclate.core.configurator as config
import nomenclate.core.exceptions as exceptions
from nomenclate.core.nomenclative import (
    InputRenderer
)
from nomenclate.core.nlog import (
    getLogger,
    DEBUG,
    INFO,
    CRITICAL
)
from nomenclate.core.tools import (
    combine_dicts
)


class TokenAttr(object):
    LOG = getLogger(__name__, level=CRITICAL)

    def __init__(self, label=None, token=None):
        try:
            self.validate_entries(label)
        except exceptions.ValidationError:
            label = None
        self.validate_entries(token)
        self.raw_string = label if label is not None else ""
        self.raw_token = token

    @property
    def token(self):
        return self.raw_token.lower()

    @token.setter
    def token(self, token):
        self.validate_entries(token)
        self.raw_token = token

    @property
    def label(self):
        return self.raw_string

    @label.setter
    def label(self, label):
        self.validate_entries(label)
        self.LOG.debug('Setting token attr %s -> %r' % (str(self), label))
        self.raw_string = label
        self.LOG.debug('%r and the raw_string is %s' % (self, self.raw_string))

    def set(self, value):
        self.label = value

    @staticmethod
    def validate_entries(*entries):
        for entry in entries:
            if isinstance(entry, (str, int)) or entry is None:
                continue
            else:
                raise exceptions.ValidationError('Invalid type %s, expected %s' % (type(entry), str))

    def __eq__(self, other):
        return self.token == other.token and self.label == other.label

    def __str__(self):
        return str(self.label)

    def __repr__(self):
        return '%s(%s):%r' % (self.token, self.raw_token, self.label)


class TokenAttrDictHandler(object):
    LOG = getLogger(__name__, level=CRITICAL)

    def __init__(self, nomenclate_object):
        self.nom = nomenclate_object
        self.LOG.info('Initializing TokenAttrDictHandler with default values %s' % self.empty_state)
        self.set_token_attrs(self.empty_state)

    @property
    def token_attrs(self):
        return self.gen_object_token_attributes(self)

    @property
    def state(self):
        return dict((name_attr.token, name_attr.label) for name_attr in self.token_attrs)

    @state.setter
    def state(self, input_dict):
        self.set_token_attrs(input_dict)

    @property
    def empty_state(self):
        self.LOG.info("Generating empty initial state from format order: %s" % self.nom.format_order)
        return {token: "" for token in self.nom.format_order}

    @property
    def unset_token_attrs(self):
        return [token_attr for token_attr in self.token_attrs if token_attr.label == '']

    @property
    def token_attr_dict(self):
        return dict([(attr.token, attr.label) for attr in self.token_attrs])

    def reset(self):
        for token_attr in self.token_attrs:
            token_attr.set('')

    def set_token_attrs(self, input_dict):
        if not input_dict:
            self.LOG.info('No changes to make, ignoring...')
            return

        self.LOG.debug('Setting token attributes %s against current state %s' % (input_dict, self.state))

        for input_attr_name, input_attr_value in iteritems(input_dict):
            self.set_token_attr(input_attr_name, input_attr_value)

        self.LOG.debug('Finished setting attributes on TokenDict %s' % [attr for attr in list(self.token_attrs)])
        self.update_nomenclate_token_attributes()

    def update_nomenclate_token_attributes(self):
        token_attrs = list(self.token_attrs)
        self.LOG.info('Updating nomenclate attributes - %s' % map(str, token_attrs))
        for token_attribute in token_attrs:
            setattr(self.nom, token_attribute.token, token_attribute)
        self.LOG.info('Finished updating Nomenclate object.')

    def set_token_attr(self, token, value):
        self.LOG.info('set_token_attr() - Setting TokenAttr %s with value %r' % (token, value))
        token_attrs = list(self.token_attrs)

        if token not in [token_attr.token for token_attr in token_attrs]:
            self.LOG.warning('Token did not exist, creating %r=%r...' % (token, value))
            self._create_token_attr(token, value)
        else:
            for token_attr in token_attrs:
                self.LOG.debug('Checking if token %r is equal to preexisting token %r' % (token, token_attr.token))
                if token == token_attr.token:
                    self.LOG.debug('Found matching token, setting value to %r' % value)
                    token_attr.label = value
                    break

    def get_token_attr(self, token):
        token_attr = getattr(self, token.lower())
        if token_attr is None:
            self.LOG.error(exceptions.SourceError('Instance has no %s token attribute set.' % token), exc_info=True)
        else:
            return token_attr

    def _create_token_attr(self, token, value):
        self.LOG.debug('_create_token_attr(%s:%s)' % (token, repr(value)))
        self.__dict__[token.lower()] = TokenAttr(label=value, token=token)

    def purge(self):
        # Should not be used in case we want to swap formats...only in the case of necessity
        self.purge_invalid_name_attrs()
        self.purge_nomenclate_excess_token_attributes()

    def purge_invalid_name_attrs(self):
        """ Removes name attrs not found in the format order
        """
        token_attrs = list(self.token_attrs)
        self.LOG.debug('Checking if current TokenAttrs %s are in format order %s' % (map(str, token_attrs),
                                                                                     self.nom.format_order))
        for token_attr in token_attrs:
            try:
                self._validate_name_in_format_order(token_attr.raw_token, self.nom.format_order)
            except exceptions.FormatError:
                self.LOG.info('Deleting invalid TokenAttr %s' % token_attr.token)
                del self.__dict__[token_attr.token]

    def purge_nomenclate_excess_token_attributes(self):
        for token_attr in list(self.gen_object_token_attributes(self.nom)):
            if token_attr.raw_token not in self.nom.format_order:
                self.LOG.info('Deleting unsynchronized Nomenclate TokenAttr %s' % token_attr.token)
                delattr(self.nom, token_attr.token)

    @staticmethod
    def gen_object_token_attributes(obj):
        for name, value in iteritems(obj.__dict__):
            if isinstance(value, TokenAttr):
                yield value

    @staticmethod
    def _validate_name_in_format_order(name, format_order):
        """ For quick checking if a key token is part of the format order
        """
        if name not in format_order:
            raise exceptions.FormatError('The name token %s is not found in the current format ordering' %
                                         format_order)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return all(map(lambda x: x[0] == x[1], zip(self.token_attrs, other.token_attrs)))
        return False

    def __getattr__(self, name):
        try:
            return self.__dict__[name]
        except KeyError:
            object.__getattribute__(self.__dict__, name)

    def __str__(self):
        return ' '.join(['%s:%s' % (token_attr.token, token_attr.label) for token_attr in self.token_attrs])


class FormatString(object):
    # Is not a hard coded (word) and does not end with any non word characters or capitals (assuming camel)
    # FORMAT_STRING_REGEX = r'(?:(?<=\()[\w]+(?=\)))|([A-Za-z0-9][^A-Z_\W]+)'
    FORMAT_STRING_REGEX = r'(?:\([\w]+\))|([A-Za-z0-9][^A-Z_\W]+)'
    SEPARATORS = '\\._-?()'

    LOG = getLogger(__name__, level=CRITICAL)

    def __init__(self, format_string=""):
        self.LOG.info('Initializing formt string with input %r' % format_string)
        self.processed_format_order = []
        self.format_string = format_string
        self.format_order = format_string
        self.swap_format(format_string)

    def swap_format(self, format_target):
        try:
            self.get_valid_format_order(format_target)
            self.format_order = format_target
            self.format_string = format_target
            self.LOG.debug('Successfully set format string: %s and format order: %s' % (self.format_string,
                                                                                        self.format_order))
        except exceptions.FormatError as e:
            msg = "Could not validate input format target %s"
            self.LOG.error('%s, %s' % (msg, e.message))

    def parse_format_order(self, format_target):
        """ Dissects the format string and gets the order of the tokens as it finds them l->r
            Splits on camel case or periods/underscores
            Modified version from this:
                http://stackoverflow.com/questions/2277352/python-split-a-string-at-uppercase-letters
        Returns [string]: list of the matching tokens
        """
        self.LOG.debug('Getting format order from target %s' % repr(format_target))
        try:
            pattern = re.compile(self.FORMAT_STRING_REGEX)
            return [match.group() for match in pattern.finditer(format_target) if None not in match.groups()]
        except TypeError:
            raise exceptions.FormatError('Format string %s is not a valid input type, must be <type str>' %
                                         format_target)

    @property
    def format_order(self):
        return self.processed_format_order

    @format_order.setter
    def format_order(self, format_target):
        if format_target:
            self.processed_format_order = self.get_valid_format_order(format_target,
                                                                      format_order=self.parse_format_order(
                                                                          format_target))
        else:
            self.processed_format_order = []

    def get_valid_format_order(self, format_target, format_order=None):
        """ Checks to see if the target format string follows the proper style
        """
        self.LOG.debug('Validating format target %r and parsing for format order %r.' % (format_target, format_order))
        format_order = format_order or self.parse_format_order(format_target)
        self.LOG.debug('Resulting format order is %s' % format_order)

        for format_str in format_order:
            format_target = re.sub(format_str, '', format_target, count=1)

        format_target = re.sub('\([\w]+\)', '', format_target)
        self.LOG.debug('After processing format_target is %s' % format_target)

        for char in format_target:
            if char not in self.SEPARATORS:
                msg = "You have specified an invalid format string %s." % format_target
                self.LOG.warning(msg)
                raise exceptions.FormatError(msg)
        return format_order

    def __str__(self):
        return str(self.format_string)


class Nomenclate(object):
    """This class deals with renaming of objects in an approved pattern
    """
    LOG = getLogger(__name__, level=CRITICAL)
    CONFIG_PATH = ['overall_config']

    NAMING_FORMAT_PATH = ['naming_formats']
    DEFAULT_FORMAT_PATH = NAMING_FORMAT_PATH + ['node', 'default']

    SUFFIXES_PATH = ['suffixes']
    OPTIONS_PATH = ['options']
    SIDE_PATH = OPTIONS_PATH + ['side']

    def __init__(self, *args, **kwargs):
        self.LOG.info('***CREATING NEW NOMENCLATE OBJECT***')
        self.cfg = config.ConfigParse()
        self.format_string_object = FormatString()

        self.CONFIG_OPTIONS = dict()

        self.reset_from_config()
        self.token_dict = TokenAttrDictHandler(self)
        self.LOG.info('Nomenclate init passed args %s and kwargs %s...processing' % (args, kwargs))
        self.merge_dict(*args, **kwargs)

    @property
    def empty_tokens(self):
        return self.token_dict.unset_token_attrs

    @property
    def tokens(self):
        return list(self.token_dict.state)

    @property
    def format_order(self):
        return self.format_string_object.format_order

    @property
    def state(self):
        return self.token_dict.state

    @state.setter
    def state(self, input_object):
        """
        Args:
            input_object Union[dict, self.__class__]: accepts a dictionary or self.__class__
        """
        self.merge_dict(input_object)

    @property
    def format(self):
        return str(self.format_string_object)

    @format.setter
    def format(self, format_target):
        self.LOG.info('Attempting to swap format to target %r' % format_target)
        orig_format, orig_order = (self.format, self.format_order)

        try:
            self.LOG.info('Looking in config for format target: %r' % format_target)
            format_target = self.cfg.get(format_target, return_type=str)
            if format_target:
                self.format_string_object.swap_format(format_target)
                self.LOG.info('Found entry: %r' % format_target)
        except exceptions.ResourceNotFoundError:
            self.LOG.info('Format target not found in config, validating as a format string...')
            self.format_string_object.swap_format(format_target)

        if hasattr(self, 'token_dict') and self.format != orig_format:
            self.LOG.info('Comparing new format order %s with old format order %s' % (self.format_order, orig_order))
            new_entries = [token for token in set(self.format_order) - set(orig_order) if not hasattr(self, token)]
            fillers = dict.fromkeys(new_entries, '')
            self.LOG.info('Format string has been changed, updating internal attributes with fillers %r' % fillers)
            self.merge_dict(fillers)

    def reset_from_config(self):
        self.initialize_config_settings()
        self.initialize_format_options()
        self.initialize_options()
        self.initialize_ui_options()

    def initialize_config_settings(self, input_dict=None):
        input_dict = input_dict or self.cfg.get(self.CONFIG_PATH, return_type=dict)
        for setting, value in iteritems(input_dict):
            setattr(self, setting, value)

    def initialize_format_options(self, format_target=''):
        """ First attempts to use format_target as a config path or gets the default format
            if it's invalid or is empty.
        Args:
            format_target Union[str, list]: can be either a query path to a format
                                            or in format of a naming string the sections should be spaced around
                                              e.g. - this_is_a_naming_string
        Returns None: raises IOError if failure
        """
        return_type = str
        self.LOG.info('initialize_format_options with format target %r' % format_target)
        try:
            self.format = format_target
            self.LOG.info('Successfully set format target.')
        except exceptions.FormatError:
            pass

        finally:
            if not format_target:
                self.LOG.debug('Format not found, replacing with default format from config path %s' %
                               self.DEFAULT_FORMAT_PATH)
                self.format_string_object.swap_format(self.cfg.get(self.DEFAULT_FORMAT_PATH, return_type=return_type))

            self.LOG.info('format target is now %s after looking in the config for a match' % format_target)

    def initialize_options(self):
        self.CONFIG_OPTIONS = self.cfg.get(self.CONFIG_PATH, return_type=dict)

    def initialize_ui_options(self):
        """
        Placeholder for all categories/sub-lists within options to be recorded here
        """
        pass

    def get(self, **kwargs):
        """Gets the string of the current name of the object
        Returns (string): the name of the object
        """
        self.LOG.info('RENDERING NOMENCLATE OBJECT')
        self.LOG.info('STATE IS: %s' % self.state)
        self.merge_dict(**kwargs)
        self.LOG.info('STATE IS: %s' % self.state)
        result = InputRenderer.render_nomenclative(self)
        return result

    def get_chain(self, end, start=None, **kwargs):
        """ Returns a list of names based on index values
        Args:
            end (int): integer for end of sequence
            start (int): optional definition of start position
        Returns (list): generated object names
        """
        # TODO: rework this entire function, very dirty function.
        """
        var_attr = self.token_dict.get_token_attr(self.VARIATION_PATH[-1])
        var_orig = var_attr.get()

        var_start, n_type = self._get_alphanumeric_index(var_orig)

        if start is None:
            start = var_start
        names = []
        for index in range(start, end + 1):
            if n_type in ['char_hi', 'char_lo']:
                capital = True if n_type == 'char_hi' else False
                var_attr.set(self._get_variation_id(index, capital))

            else:
                var_attr.set(str(index))
            if 'var' in kwargs:
                kwargs.pop("var", None)
            names.append(self.get(**kwargs))
        var_attr.set(var_orig)
        return names
        """
        raise NotImplementedError

    def merge_dict(self, *args, **kwargs):
        if args or kwargs:
            input_dict = self._convert_input(*args, **kwargs)
            self._sift_and_init_configs(input_dict)
            self.LOG.info('Done sifting, now input is %s' % input_dict)
            self.token_dict.state = input_dict
        else:
            self.LOG.warning('Nothing to update...empty args/kwargs...skipping.')

    def _convert_input(self, *args, **kwargs):
        self.LOG.info('Converting input args %s and kwargs %s' % (args, kwargs))
        args = [arg.state if isinstance(arg, Nomenclate) else arg for arg in args]
        self.LOG.info('Args after Nomenclate converstion: %s' % args)
        input_dict = combine_dicts(*args, **kwargs)
        self.LOG.info('Converted to %s' % input_dict)
        return input_dict

    def get_token_settings(self, token, default=None):
        setting_dict = {}
        for key, value in iteritems(self.__dict__):
            if '%s_' % token in key and not callable(key) and not isinstance(value, TokenAttr):
                setting_dict[key] = self.__dict__.get(key, default)
        return setting_dict

    def _sift_and_init_configs(self, input_dict):
        """ Removes all key/v for keys that exist in the overall config and activates them.
            Used to weed out config keys from tokens in a given input.
        """
        self.LOG.info('_sift_and_init_configs() - removing config settings from input %s' % input_dict)
        configs = {}
        for k, v in iteritems(input_dict):
            if (k not in map(str.lower, self.format_order) and
                    any([f_order.lower() in k for f_order in self.format_order])):
                try:
                    self.cfg.get(self.CONFIG_PATH + [k])
                except exceptions.ResourceNotFoundError:
                    pass
                finally:
                    configs[k] = v

        for key, val in iteritems(configs):
            self.LOG.info('Sifting out found config setting %s' % key)
            input_dict.pop(key, None)

        self.initialize_config_settings(input_dict=configs)

    def __eq__(self, other):
        return self.token_dict == other.token_dict

    def __str__(self):
        return self.get()

    def __setattr__(self, key, value):
        attr = getattr(self, key, None)
        if isinstance(attr, TokenAttr):
            attr.label = value.label if isinstance(value, TokenAttr) else value
            self.LOG.debug('User setting TokenAttr %s to %s, now is %s' % (attr, value, attr.label))
        else:
            self.LOG.debug('User setting attribute %s to %s' % (key, value))
            object.__setattr__(self, key, value)

    def __dir__(self):
        return dir(super(Nomenclate, self)) + self.tokens



# Ensure Python 2/3 compatibility: http://python-future.org/compatible_idioms.html
from __future__ import print_function
from future.utils import iteritems
import unittest
import re
from pprint import pprint
import nomenclate.core.nomenclature as nm
import nomenclate.core.configurator as config
import nomenclate.core.exceptions as exceptions


class TestBase(unittest.TestCase):
    def setUp(self):
        self.fixtures = []
        print('running testBase setup!')

    def tearDown(self):
        for fixture in self.fixtures:
            del fixture

    @staticmethod
    def checkEqual(L1, L2):
        return len(L1) == len(L2) and sorted(L1) == sorted(L2)


class TestTokenAttrBase(TestBase):
    def setUp(self):
        super(TestTokenAttrBase, self).setUp()
        self.token_attr = nm.TokenAttr('test_label', 'test_token')
        self.fixtures.append(self.token_attr)


class TestTokenAttrInstantiate(TestTokenAttrBase):
    def test_empty_instantiate(self):
        self.assertEquals(nm.TokenAttr().label, '')

    def test_valid_instantiate(self):
        self.fixtures.append(nm.TokenAttr('test', 'test'))

    def test_invalid_instantiate_label(self):
        self.assertRaises(exceptions.ValidationError, nm.TokenAttr, 'test', [])

    def test_invalid_instantiate_token(self):
        self.assertRaises(exceptions.ValidationError, nm.TokenAttr, {}, 'test')

    def test_state(self):
        self.assertEquals(self.token_attr.label, 'test_label')
        self.assertEquals(self.token_attr.token, 'test_token')


class TestTokenAttrSet(TestTokenAttrBase):
    def test_set_invalid(self):
        with self.assertRaises(exceptions.ValidationError):
            nm.TokenAttr().label = [1]
        with self.assertRaises(exceptions.ValidationError):
            nm.TokenAttr().label = {1: 1}


class TestTokenAttrGet(TestTokenAttrBase):
    def test_get(self):
        self.assertEquals(self.token_attr.label, 'test_label')

    def test_get_empty(self):
        self.assertEquals(nm.TokenAttr().label, '')


class TestNomenclateBase(TestBase):
    def setUp(self):
        super(TestBase, self).setUp()
        self.cfg = config.ConfigParse()
        self.test_format = 'side_location_nameDecoratorVar_childtype_purpose_type'
        self.test_format_b = 'location_nameDecoratorVar_childtype_purpose_type_side'

        self.nom = nm.Nomenclate()

        # Inject our fake config
        self.nom.cfg = self.cfg
        self.nom.side.set('left')
        self.nom.name.set('testObject')
        self.nom.type.set('locator')
        self.nom.var.set('A')
        self.fixtures = [self.cfg, self.nom, self.test_format_b, self.test_format]


class TestNomenclateTokens(TestNomenclateBase):
    pass


class TestNomenclateState(TestNomenclateBase):
    def test_state_clear(self):
        previous_state = self.nom.state
        self.nom.token_dict.reset()
        self.assertEquals(self.nom.state,
                          {'location': '', 'type': '', 'name': '', 'side': '', 'var': '', 'purpose': '',
                           'decorator': '', 'childtype': ''})
        self.nom.state = previous_state

    @unittest.skip
    def test_state_purge(self):
        previous_state = self.nom.state
        self.nom.token_dict.purge_name_attrs()
        self.assertEquals(self.nom.state, {})
        self.nom.state = previous_state

    def test_state_valid(self):
        self.assertEquals(self.nom.state,
                          {'childtype': '',
                           'decorator': '',
                           'location': '',
                           'name': 'testObject',
                           'purpose': '',
                           'side': 'left',
                           'type': 'locator',
                           'var': 'A'})


class TestNomenclateResetFromConfig(TestNomenclateBase):
    def test_refresh(self):
        self.assertIsNone(self.nom.reset_from_config())


class TestNomenclateInitializeConfigSettings(TestNomenclateBase):
    pass


class TestNomenclateInitializeFormatOptions(TestNomenclateBase):
    def test_switch_naming_format_from_str(self):
        self.nom.initialize_format_options(self.test_format_b)
        self.assertTrue(self.checkEqual(self.nom.format_order,
                                        ['side', 'location', 'name', 'Decorator', 'Var', 'childtype', 'purpose',
                                         'type']))

        self.nom.initialize_format_options(self.test_format)
        self.assertTrue(self.checkEqual(self.nom.format_order,
                                        ['side', 'location', 'name', 'Decorator', 'Var', 'childtype', 'purpose',
                                         'type']))

    def test_switch_naming_format_from_config(self):
        self.nom.initialize_format_options(['naming_formats', 'node', 'format_lee'])
        print(self.nom.format_order)
        self.assertTrue(self.checkEqual(self.nom.format_order,
                                        ['type', 'childtype', 'space', 'purpose', 'name', 'side']))
        self.nom.initialize_format_options(self.test_format)


class TestNomenclateInitializeOptions(TestNomenclateBase):
    def test_options_stored(self):
        self.nom.CONFIG_OPTIONS = None
        self.nom.initialize_options()
        self.assertIsNotNone(self.nom.CONFIG_OPTIONS)


class TestNomenclateInitializeUiOptions(TestNomenclateBase):
    def test_pass_through(self):
        self.nom.initialize_ui_options()


class TestNomenclateMergeDict(TestNomenclateBase):
    pass


class TestNomenclateGetFormatOrderFromFormatString(TestNomenclateBase):
    def test_get_format_order(self):
        self.assertEquals(self.nom.format_string_object.parse_format_order(self.test_format),
                          ['side', 'location', 'name', 'Decorator', 'Var', 'childtype', 'purpose', 'type'])


class TestNomenclateGet(TestNomenclateBase):
    def test_get(self):
        print(self.nom.state)
        self.assertEquals(self.nom.get(), 'l_testObjectA_LOC')

    def test_get_after_change(self):
        previous_state = self.nom.state
        self.nom.location.set('rear')
        self.assertEquals(self.nom.get(), 'l_rr_testObjectA_LOC')
        self.nom.state = previous_state


class TestNomenclateGetChain(TestNomenclateBase):
    def test_get_chain(self):
        self.assertRaises(NotImplementedError, self.nom.get_chain, (0, 5))
        #self.assertIsNone(self.nom.get_chain(0, 5))


class TestNomenclateUpdateTokenAttributes(TestNomenclateBase):
    pass


class TestNomenclateComposeName(TestNomenclateBase):
    pass


class TestNomenclateEq(TestNomenclateBase):
    def test_equal(self):
        other = nm.Nomenclate(self.nom)
        self.assertTrue(other == self.nom)

    def test_inequal_one_diff(self):
        other = nm.Nomenclate(self.nom.state)
        other.name = 'ronald'
        self.assertFalse(other == self.nom)

    def test_inequal_multi_diff(self):
        other = nm.Nomenclate(self.nom.state)
        other.name = 'ronald'
        other.var = 'C'
        other.type = 'joint'
        self.assertFalse(other == self.nom)


class TestNomenclateRepr(TestNomenclateBase):
    def test__str__(self):
        self.assertEquals(str(self.nom), 'l_testObjectA_LOC')


class TestFormatStringBase(TestBase):
    def setUp(self):
        super(TestFormatStringBase, self).setUp()
        self.fs = nm.FormatString()
        self.fixtures.append(self.fs)


class TestFormatStringValidateFormatString(TestFormatStringBase):
    def test_get__validate_format_string_valid(self):
        self.fs.get_valid_format_order('side_mide')

    def test_get__validate_format_string__is_format_invalid(self):
        print(repr(self))
        self.assertRaises(exceptions.FormatError, self.fs.get_valid_format_order, 'notside;blah')



#!/usr/bin/env python
# Ensure Python 2/3 compatibility: http://python-future.org/compatible_idioms.html
from __future__ import print_function
from future.utils import iteritems
from six import add_metaclass
import re
import string
import datetime
import dateutil.parser as p
import nomenclate.core.exceptions as exceptions
from nomenclate.core.nlog import (
    getLogger,
    DEBUG,
    CRITICAL,
    INFO
)
from nomenclate.core.tools import (
    gen_dict_key_matches,
    get_keys_containing,
    flatten
)


class TokenMatch(object):
    LOG = getLogger(__name__, level=CRITICAL)

    def __init__(self, regex_match, substitution, group_name='token'):
        self.match = regex_match.group(group_name)
        self.start = regex_match.start(group_name)
        self.end = regex_match.end(group_name)
        self.sub = str(substitution)

    @property
    def span(self):
        return self.end - self.start

    def adjust_position(self, other, adjust_by_sub_delta=True):
        self._validate_adjuster(other)
        adjustment = other.span - len(other.sub) if adjust_by_sub_delta else other.span
        if adjustment:
            self.LOG.debug('Adjusting %s by %d' % (self, adjustment))
            self._adjust_order(adjustment)

    def _validate_adjuster(self, other):
        if not isinstance(other, self.__class__):
            raise IOError('Only TokenMatch objects are valid inputs to adjust_position')

        other.overlaps(self)

        if self < other:
            raise IndexError('Current TokenMatch is not affected by input TokenMatch\n\t%s\n\t%s' % (repr(self),
                                                                                                     repr(other)))

    def _adjust_order(self, adjust_value):
        self.start -= adjust_value
        self.end -= adjust_value
        self.LOG.debug('New positions are (%d-%d)' % (self.start, self.end))

    def overlaps(self, other):
        if self in other or other in self:
            raise exceptions.OverlapError('Match %s overlaps with an existing match:\n\t%s\n\t%s' % (self.match,
                                                                                                     self, other))
        return True

    def __contains__(self, other):
        try:
            return (self.start < other.start < self.end or self.start < other.end < self.end)
        except:
            raise NotImplementedError(
                '{C} objects do not handle in syntax for non class objects'.format(C=self.__class__.__name__))

    def __eq__(self, other):
        try:
            return (self.start == other.start and self.end == other.end and
                    self.match == other.match and self.sub == other.sub)
        except:
            raise NotImplementedError(
                '{C} objects do not handle == syntax for non class objects'.format(C=self.__class__.__name__))

    def __lt__(self, other):
        if isinstance(other, self.__class__):
            return self.end <= other.start
        else:
            raise NotImplementedError

    def __gt__(self, other):
        if isinstance(other, self.__class__):
            return self.start >= other.end
        else:
            raise NotImplementedError

    def __str__(self):
        return '%s (%d)- [%d:%d] - replacement = %s' % (self.match, self.span, self.start, self.end, self.sub)


class Nomenclative(object):
    LOG = getLogger(__name__, level=CRITICAL)

    def __init__(self, input_str):
        self.str = input_str
        self.token_matches = []

    def process_matches(self):
        build_str = self.str
        for token_match in self.token_matches:
            if token_match.match == build_str[token_match.start:token_match.end]:
                self.LOG.debug('Processing: %s - %s - %s\n\t%s' % (token_match.match,
                                                                   token_match.sub,
                                                                   token_match.sub,
                                                                   build_str))

                build_str = build_str[:token_match.start] + token_match.sub + build_str[token_match.end:]
                self.adjust_other_matches(token_match)
                self.LOG.debug('Processed as:\n\t%s' % build_str)
        return build_str

    def adjust_other_matches(self, adjuster_match):
        for token_match in [token_match for token_match in self.token_matches if token_match != adjuster_match]:
            try:
                token_match.adjust_position(adjuster_match)
            except IndexError:
                pass
        adjuster_match.end = adjuster_match.start + len(adjuster_match.sub)
        adjuster_match.match = adjuster_match.sub

    def add_match(self, regex_match, substitution):
        token_match = TokenMatch(regex_match, substitution)
        try:
            self.validate_match(token_match)
            self.token_matches.append(token_match)
            self.LOG.info('Added match %s' % self.token_matches[-1])
        except IndexError:
            self.LOG.warning('Not adding match %s as it conflicts with a preexisting match' % token_match)

    def validate_match(self, token_match_candidate):
        for token_match in self.token_matches:
            try:
                token_match.overlaps(token_match_candidate)
            except exceptions.OverlapError as e:
                self.LOG.error(e.message)

    def __str__(self):
        matches = '' if not self.token_matches else '\n'.join(map(str, self.token_matches))
        return '%s:%s' % (self.str, matches)


class InputRenderer(type):
    RENDER_FUNCTIONS = {}
    REGEX_PARENTHESIS = r'([\(\)]+)'
    REGEX_BRACKETS = r'([\{\}]+)'
    REGEX_STATIC_TOKEN = r'(\(\w+\))'
    REGEX_BRACKET_TOKEN = r'(\{\w+\})'
    REGEX_TOKEN_SEARCH = r'(?P<token>((?<![a-z]){TOKEN}(?![0-9]))|((?<=[a-z]){TOKEN_CAPITALIZED}(?![0-9])))'

    LOG = getLogger(__name__, level=CRITICAL)

    def __new__(mcs, name, bases, dct):
        cls = type.__new__(mcs, name, bases, dct)
        token = dct.get('token', None)
        if token:
            mcs.RENDER_FUNCTIONS[token] = cls
        return cls

    @classmethod
    def render_unique_tokens(cls, nomenclate_object, input_dict):
        cls.LOG.info('current list of render functions: %s' % list(cls.RENDER_FUNCTIONS))
        for k, v in iteritems(input_dict):
            cls.LOG.info('Checking for unique token on token %s:%r' % (k, v))
            if v:
                for func in [func for func in list(cls.RENDER_FUNCTIONS)
                             if k.replace(func, '').isdigit() or k.replace(func, '') == '']:
                    renderer = cls.RENDER_FUNCTIONS.get(func, None)
                    cls.LOG.info('Attempting to find token unique render function for token %s with renderer %s' % (k, renderer))
                    if 'render' in dir(renderer):
                        cls.LOG.info('render_unique_tokens() - token %s renderer %s with token settings %s' %
                                     (k, v, nomenclate_object.get_token_settings(k)))

                        rendered_token = renderer.render(v, k, nomenclate_object, **nomenclate_object.get_token_settings(k))
                        cls.LOG.info('Unique token %s rendered as: %s' % (k, rendered_token))

                        input_dict[k] = rendered_token

    @classmethod
    def render_nomenclative(cls, nomenclate_object):
        nomenclative = Nomenclative(nomenclate_object.format)
        token_values = nomenclate_object.token_dict.token_attr_dict
        cls.LOG.info('render_nomenclative() - Current state is %s with nomenclative %s' % (token_values, nomenclative))
        cls.render_unique_tokens(nomenclate_object, token_values)
        rendered_nomenclative = nomenclate_object.format
        cls.LOG.info('Finished rendering unique tokens.')
        cls._prepend_token_match_objects(token_values, rendered_nomenclative)

        for token, match_value in iteritems(token_values):
            nomenclative.add_match(*match_value)

        cls.LOG.info('Before processing state has been updated to %s' % nomenclative)
        rendered_nomenclative = cls.cleanup_formatted_string(nomenclative.process_matches())
        cls.LOG.info('Finally converted to %s' % rendered_nomenclative)
        return rendered_nomenclative

    @classmethod
    def _prepend_token_match_objects(cls, token_values, incomplete_nomenclative):
        for token, value in iteritems(token_values):
            re_token = cls.REGEX_TOKEN_SEARCH.format(TOKEN=token,
                                                     TOKEN_CAPITALIZED=token[0].upper()+token[1:])
            re_matches = re.finditer(re_token, incomplete_nomenclative, 0)

            for re_match in re_matches:
                token_values[token] = (re_match, value)

        cls._clear_non_matches(token_values)

    @staticmethod
    def _clear_non_matches(token_values):
        to_delete = []
        for token, value in iteritems(token_values):
            if isinstance(value, str) or not isinstance(value, tuple):
                to_delete.append(token)

        for delete in to_delete:
            token_values.pop(delete)

    @staticmethod
    def _render_replacements(replacements_dict, incomplete_nomenclative):
        for token, match_value in iteritems(replacements_dict):
            match, value = match_value
            incomplete_nomenclative = incomplete_nomenclative.replace(match, value)

    @classmethod
    def cleanup_formatted_string(cls, formatted_string):
        """ Removes unused tokens/removes surrounding and double underscores
        Args:
            formatted_string (string): string that has had tokens replaced
        Returns (string): cleaned up name of object
        """
        # TODO: chunk this out to sub-processes for easier error checking, could be own class
        # Remove whitespace
        result = formatted_string.replace(' ', '')
        # Remove any static token parentheses
        result = re.sub(cls.REGEX_PARENTHESIS, '', result)
        # Remove any multiple underscores
        result = re.sub('_+', '_', result)
        # Remove trailing or preceding non letter characters
        result = re.sub(r'(^[\W_]+)|([\W_]+$)', '', result)
        #  not sure what this one was...but certainly not it.
        result = re.sub(r'(\()|(\))', '', result)
        return result

    @staticmethod
    def _get_alphanumeric_index(query_string):
        """ Given an input string of either int or char, returns what index in the alphabet and case it is
        Args:
            query_string (str): query string
        Returns [int, str]: list of the index and type
        """
        #TODO: could probably rework this. it works, but it's ugly as hell.
        try:
            return [int(query_string), 'int']
        except ValueError:
            if len(query_string) == 1:
                if query_string.isupper():
                    return [string.ascii_uppercase.index(query_string), 'char_hi']
                elif query_string.islower():
                    return [string.ascii_lowercase.index(query_string), 'char_lo']
            else:
                raise IOError('The input is a string longer than one character')
        return [0, 'char_hi']


@add_metaclass(InputRenderer)
class RenderBase(object):
    LOG = getLogger(__name__, level=CRITICAL)
    #metaclass__ = InputRenderer
    token = None

    @classmethod
    def render(cls, arg, token, nomenclate_object, **kwargs):
        raise NotImplementedError

    @classmethod
    def get_config_match(cls,
                         query_string,
                         token,
                         entry_path,
                         return_type,
                         nomenclate_object,
                         config_entry_suffix='length',
                         **kwargs):

        try:
            options = nomenclate_object.cfg.get(entry_path, return_type=return_type)

            if type(options) == dict:
                options = list(gen_dict_key_matches(query_string, options))
            options = list(flatten(options))

            criteria = kwargs.get('%s_%s' % (cls.token, config_entry_suffix), None)
            cls.LOG.debug('%s(%s) options: %s criteria: %s' % (token, query_string, options, criteria))

            for option in options:
                cls.LOG.debug('Running through option %s' % option)
                if len(option) == criteria and criteria:
                    cls.LOG.debug('Found item matching criteria: %s -> %s' % (criteria, option))
                    return option

            try:
                result_option = next(options)
            except (StopIteration, TypeError):
                result_option = options

            cls.LOG.debug('Best match is: %s' % result_option)
            return result_option

        except exceptions.ResourceNotFoundError:
            cls.LOG.warning('No entry for token %s - defaulting to current: %s' % (token,
                                                                                   query_string))
            return query_string


class RenderDate(RenderBase):
    token = 'date'

    @classmethod
    def render(cls, date, token, nomenclate_object, **kwargs):
        if date == 'now':
            d = datetime.datetime.now()
        else:
            try:
                d = p.parse(date)
            except ValueError:
                return ''
        date_format = getattr(nomenclate_object, '%s_format' % cls.token, '%Y-%m-%d')
        # return d.strftime(get_keys_containing('format', nomenclate_object.__dict__, '%Y-%m-%d'))
        return d.strftime(date_format)


class RenderVar(RenderBase):
    __metaclass__ = InputRenderer
    token = 'var'

    @classmethod
    def render(cls, var, token, nomenclate_object, **kwargs):
        if var:
            var_index = kwargs.get('%s_index' % cls.token, 0)
            return cls._get_variation_id(var_index, var.isupper())
        else:
            return var

    @staticmethod
    def _get_variation_id(integer, capital=False):
        """ Convert an integer value to a character. a-z then double aa-zz etc
        Args:
            integer (int): integer index we're looking up
            capital (bool): whether we convert to capitals or not
        Returns (str): alphanumeric representation of the index
        """
        # calculate number of characters required
        base_power = base_start = base_end = 0
        while integer >= base_end:
            base_power += 1
            base_start = base_end
            base_end += pow(26, base_power)
        base_index = integer - base_start

        # create alpha representation
        alphas = ['a'] * base_power
        for index in range(base_power - 1, -1, -1):
            alphas[index] = chr(int(97 + (base_index % 26)))
            base_index /= 26

        characters = ''.join(alphas)
        return characters.upper() if capital else characters


class RenderVersion(RenderBase):
    __metaclass__ = InputRenderer
    token = 'version'

    @classmethod
    def render(cls, version, token, nomenclate_object, **kwargs):
        padding = kwargs.get('%s_padding' % token, 4)
        format = kwargs.get('%s_format' % token, '#')
        version_string = format.replace('#', '%0{0}d')
        return version_string.format(padding) % version


class RenderType(RenderBase):
    __metaclass__ = InputRenderer
    token = 'type'

    @classmethod
    def render(cls, ttype, token, nomenclate_object, **kwargs):
        return cls.get_config_match(ttype,
                                    cls.token,
                                    nomenclate_object.SUFFIXES_PATH,
                                    dict,
                                    nomenclate_object,
                                    **kwargs) or ttype


class RenderSide(RenderBase):
    __metaclass__ = InputRenderer
    token = 'side'

    @classmethod
    def render(cls, side, token, nomenclate_object, **kwargs):

        return cls.get_config_match(side,
                                    cls.token,
                                    nomenclate_object.OPTIONS_PATH + [cls.token, side],
                                    list,
                                    nomenclate_object,
                                    **kwargs)


class RenderLocation(RenderBase):
    __metaclass__ = InputRenderer
    token = 'location'

    @classmethod
    def render(cls, location, token, nomenclate_object, **kwargs):

        return cls.get_config_match(location,
                                    cls.token,
                                    nomenclate_object.OPTIONS_PATH + [cls.token, location],
                                    list,
                                    nomenclate_object,
                                    **kwargs)


class RenderDiscipline(RenderBase):
    __metaclass__ = InputRenderer
    token = 'discipline'

    @classmethod
    def render(cls, discipline, token, nomenclate_object, **kwargs):

        return cls.get_config_match(discipline,
                                    cls.token,
                                    nomenclate_object.OPTIONS_PATH + [cls.token, discipline],
                                    list,
                                    nomenclate_object,
                                    **kwargs)



import sys

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand


# ref: https://tox.readthedocs.org/en/latest/example/basic.html#integration-with-setuptools-distribute-test-commands
class Tox(TestCommand):
    user_options = [('tox-args=', 'a', "Arguments to pass to tox")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.tox_args = ''

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import tox
        import shlex
        errno = tox.cmdline(args=shlex.split(self.tox_args))
        sys.exit(errno)


class ToxWithRecreate(Tox):
    description = ('Run tests, but recreate the testing environments first. '
                   '(Useful if the test dependencies change.)')

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.tox_args = '-r'


setup(
    name='nomenclate',
    version='1.0.6',
    packages=find_packages(),
    package_data={'configYML': ['nomenclate/core/*.yml']},
    include_package_data=True,
    url='https://github.com/andresmweber/nomenclate',
    license='MIT',
    author='Andres Weber',
    description='Naming Convention Generator ',
    long_description='Tool for generating string-based labels based on a preset convention configuration.',
    keywords='naming conventions labels config convention name',
    
    entry_points={
        'console_scripts':[
            'nomenclate = nomenclate.app:run',
        ]
    },
    
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Multimedia :: Graphics :: 3D Modeling',
    ],

    install_requires=[
        'configparser',
        'future',
        'python-dateutil',
        'PyYAML',
        'six'
    ],

    # tox is responsible for setting up the test runner and its dependencies
    # (e.g., code coverage tools) -- see the tox.ini file
    tests_require=[
        'python-mimeparse',
        'pbr',
        'mox3',
        'virtualenv',
        'traceback2',
        'coverage',
        'extras',
        'funcsigs',
        'linecache2',
        'pluggy',
        'py',
        'fixtures',
        'unittest2',
        'mock',
        'nose',
        'pyfakefs',
        'tox >= 1.9, < 3',
        'testtools',
        'tox'
    ],

    cmdclass={
        'test': Tox,
        'clean_test': ToxWithRecreate,
    },
)
